const vscode = require('vscode');
const path = require('path');
const fs = require('fs');
const http = require('http');

// store name
let _storeName = null;

let statusBarItem; // Reference to the status bar item

function createStatusBarButton() {
    if (!statusBarItem) {
        // Create a new Status Bar Item
        statusBarItem = vscode.window.createStatusBarItem(vscode.StatusBarAlignment.Right, 100);
        statusBarItem.text = "$(play) Run Query"; // Button label with an icon
        statusBarItem.tooltip = "Send the current JavaScript file to the server";
        statusBarItem.command = "graphlake.sendQueryToServer"; // Command to execute
    }
}

const CONFIG_FILE_NAME = "graphlake.json";

function getConfigFilePath() {
    const workspaceFolders = vscode.workspace.workspaceFolders;
    if (!workspaceFolders || workspaceFolders.length === 0) {
        vscode.window.showErrorMessage("Please open a folder to use the GraphLake extension.");
        return null;
    }
    return path.join(workspaceFolders[0].uri.fsPath, CONFIG_FILE_NAME);
}

function readConfig() {
    const configFilePath = getConfigFilePath();
    if (!configFilePath || !fs.existsSync(configFilePath)) {
        return {};
    }

    const workspaceFolders = vscode.workspace.workspaceFolders;
    
    // create folder for data, if it does not exist
    const dataFolder = path.join(workspaceFolders[0].uri.fsPath, "data");
    if (!fs.existsSync(dataFolder)) {
        fs.mkdirSync(dataFolder);
    }

    // create folder for queries, if it does not exist
    const queriesFolder = path.join(workspaceFolders[0].uri.fsPath, "queries");
    if (!fs.existsSync(queriesFolder)) {
        fs.mkdirSync(queriesFolder);
    }

    // create folder for schemas, if it does not exist
    const schemasFolder = path.join(workspaceFolders[0].uri.fsPath, "schemas");
    if (!fs.existsSync(schemasFolder)) {
        fs.mkdirSync(schemasFolder);
    }


    return JSON.parse(fs.readFileSync(configFilePath, "utf-8"));
}

class StoresProvider {
    constructor(graphsProvider) {
        this.graphsProvider = graphsProvider;
        this.stores = [];
        this.loadStores();
        this._onDidChangeTreeData = new vscode.EventEmitter();
        this.onDidChangeTreeData = this._onDidChangeTreeData.event;
    }

    refresh() {
        this._onDidChangeTreeData.fire();
    }

    getTreeItem(storeName) {
        const treeItem = new vscode.TreeItem(storeName, vscode.TreeItemCollapsibleState.None);
        treeItem.contextValue = "store";
        
        treeItem.iconPath = new vscode.ThemeIcon("database");

        treeItem.command = {
            command: "graphlake.selectStore",
            title: "Select Store",
            arguments: [storeName],
        };
        return treeItem;
    }

    getChildren() {
        return this.stores;
    }

    async fetchStoresFromServer() {
        return new Promise((resolve, reject) => {
            const options = {
                hostname: server,
                port: port,
                path: "/stores",
                method: "GET",
            };

            const req = http.request(options, (res) => {
                let data = "";

                res.on("data", (chunk) => {
                    data += chunk;
                });

                res.on("end", () => {
                    if (res.statusCode === 200) {
                        try {
                            const stores = JSON.parse(data);
                            resolve(stores);
                        } catch (error) {
                            reject(new Error("Failed to parse server response."));
                        }
                    } else {
                        reject(new Error(`Server returned status code ${res.statusCode}`));
                    }
                });
            });

            req.on("error", (error) => {
                reject(error);
            });

            req.end();
        });
    }

    async addStoreToServer(storeName) {
        return new Promise((resolve, reject) => {
            const options = {
                hostname: server,
                port: port,
                path: `/stores/${encodeURIComponent(storeName)}`,
                method: "POST",
            };

            const req = http.request(options, (res) => {
                let data = "";

                res.on("data", (chunk) => {
                    data += chunk;
                });

                res.on("end", () => {
                    if (res.statusCode === 200) {
                        resolve(`Store "${storeName}" created successfully.`);
                    } else {
                        reject(new Error(`Failed to create store. Server returned ${res.statusCode}: ${data}`));
                    }
                });
            });

            req.on("error", (error) => {
                reject(error);
            });

            req.end();
        });
    }

    async loadStores() {
        vscode.window.withProgress(
            {
                location: vscode.ProgressLocation.Window,
                title: "Fetching stores from server...",
            },
            async () => {
                try {
                    const stores = await this.fetchStoresFromServer();
                    this.stores = stores || [];
                    this.refresh();
                } catch (error) {
                    vscode.window.showErrorMessage(`Failed to load stores: ${error.message}`);
                }
            }
        );
    }

    async addStore(storeName) {
        vscode.window.withProgress(
            {
                location: vscode.ProgressLocation.Notification,
                title: `Creating store "${storeName}"...`,
            },
            async () => {
                try {
                    const message = await this.addStoreToServer(storeName);
                    vscode.window.showInformationMessage(message);
                    await this.loadStores(); // Refresh the list after successfully creating the store
                } catch (error) {
                    vscode.window.showErrorMessage(`Failed to create store: ${error.message}`);
                }
            }
        );
    }

    deleteStore(storeName) {
        vscode.window.showWarningMessage("Cannot delete stores manually. Use the server's API.");
    }
}

// Mock function to simulate fetching graphs from a server:
function fetchGraphsFromServer() {
    return ["graphA", "graphB", "graphC"];
}


let _graphName = null;

class GraphsProvider {
    constructor() {
        this.graphs = [];
        this._onDidChangeTreeData = new vscode.EventEmitter();
        this.onDidChangeTreeData = this._onDidChangeTreeData.event;

        // Immediately fetch graphs for the active store when constructed
        this.loadGraphs();
    }

    /**
     * Fetch the list of graphs from the server.
     * Updates this.graphs and refreshes the tree.
     */
    async loadGraphs() {
        if (_storeName === null) {
            vscode.window.showErrorMessage("No store selected.");
            return;
        }

        const path = `/stores/${encodeURIComponent(_storeName)}/graphs`;
        try {
            const data = await this._httpRequest("GET", path);
            const parsed = JSON.parse(data);
            this.graphs = Array.isArray(parsed) ? parsed : [];
        } catch (error) {
            vscode.window.showErrorMessage(`Failed to load graphs: ${error.message}`);
            this.graphs = [];
        }
        this.refresh();
    }

    /**
     * Create a new graph on the server, then reload the list.
     */
    async createGraph(graphName) {
        const path = `/stores/${encodeURIComponent(_storeName)}/graphs/${encodeURIComponent(graphName)}`;
        try {
            await this._httpRequest("POST", path);
            vscode.window.showInformationMessage(`Graph "${graphName}" created successfully.`);
        } catch (error) {
            vscode.window.showErrorMessage(`Failed to create graph: ${error.message}`);
        }
        this.loadGraphs();
    }

    /**
     * Delete an existing graph, then reload the list.
     */
    async deleteGraph(graphName) {
        const path = `/stores/${encodeURIComponent(_storeName)}/graphs/${encodeURIComponent(graphName)}`;
        try {
            await this._httpRequest("DELETE", path);
            vscode.window.showInformationMessage(`Graph "${graphName}" deleted successfully.`);
        } catch (error) {
            vscode.window.showErrorMessage(`Failed to delete graph: ${error.message}`);
        }
        this.loadGraphs();
    }

    /**
     * Refresh the tree items in VS Code.
     */
    refresh() {
        this._onDidChangeTreeData.fire();
    }

    /**
     * Standard TreeDataProvider methods.
     */
    getChildren() {
        return this.graphs;
    }

    getTreeItem(graphName) {
        const treeItem = new vscode.TreeItem(graphName, vscode.TreeItemCollapsibleState.None);

        // Attach a command to each graph node so we can select it
        treeItem.command = {
            command: "graphlake.selectGraph",
            title: "Select Graph",
            arguments: [graphName], // We'll receive graphName in the command callback
        };
    
        return treeItem;
    }

    /**
     * Internal helper to make an HTTP request and return the response body.
     */
    _httpRequest(method, path) {
        return new Promise((resolve, reject) => {
            const options = {
                hostname: server,
                port: port,
                path,
                method,
            };

            const req = http.request(options, (res) => {
                let data = "";
                res.on("data", (chunk) => {
                    data += chunk;
                });
                res.on("end", () => {
                    if (res.statusCode >= 200 && res.statusCode < 300) {
                        resolve(data);
                    } else {
                        reject(
                            new Error(
                                `Server returned ${res.statusCode}: ${data}`
                            )
                        );
                    }
                });
            });

            req.on("error", (error) => {
                reject(error);
            });

            req.end();
        });
    }
}

class QueriesProvider {
    constructor() {
        this._onDidChangeTreeData = new vscode.EventEmitter();
        this.onDidChangeTreeData = this._onDidChangeTreeData.event;
    }

    refresh() {
        this._onDidChangeTreeData.fire(); // Triggers a refresh for the view
    }

    getTreeItem(element) {
        const fileName = path.basename(element);
        const treeItem = new vscode.TreeItem(vscode.Uri.file(element), vscode.TreeItemCollapsibleState.None);
        treeItem.label = fileName;
        treeItem.command = {
            command: 'vscode.open',
            title: 'Open File',
            arguments: [vscode.Uri.file(element)]
        };
        return treeItem;
    }

    async getChildren(element) {
        if (!element) {
            // Find all .js files in the current workspace
            const workspaceFolders = vscode.workspace.workspaceFolders;
            if (!workspaceFolders || workspaceFolders.length === 0) {
                return [];
            }
            const pattern = new vscode.RelativePattern(workspaceFolders[0], 'queries/**/*.{ru,rq,js}');
            const files = await vscode.workspace.findFiles(pattern);
            return files.map(f => f.fsPath);
        }
        return [];
    }
}

class DataProvider {
    constructor() {
        this._onDidChangeTreeData = new vscode.EventEmitter();
        this.onDidChangeTreeData = this._onDidChangeTreeData.event;
    }

    refresh() {
        this._onDidChangeTreeData.fire(); // Triggers a refresh for the view
    }

    getTreeItem(element) {
        const fileName = path.basename(element);
        const treeItem = new vscode.TreeItem(vscode.Uri.file(element), vscode.TreeItemCollapsibleState.None);
        treeItem.label = fileName;
        treeItem.command = {
            command: 'vscode.open',
            title: 'Open File',
            arguments: [vscode.Uri.file(element)]
        };
        return treeItem;
    }

    async getChildren(element) {
        if (!element) {
            // Find all .js files in the current workspace
            const workspaceFolders = vscode.workspace.workspaceFolders;
            if (!workspaceFolders || workspaceFolders.length === 0) {
                return [];
            }
            const pattern = new vscode.RelativePattern(workspaceFolders[0], 'data/**/*.nt');
            const files = await vscode.workspace.findFiles(pattern);
            return files.map(f => f.fsPath);
        }
        return [];
    }
}

function sendDataToServer(content) {
    return new Promise((resolve, reject) => {
        const options = {
            hostname: server,
            port: port,
            path:  '/stores/' + _storeName + '/graphs/' + _graphName + '/import',
            method: 'POST',
            headers: {
                'Content-Type': 'application/javascript',
                'Content-Length': Buffer.byteLength(content)
            }
        };

        const req = http.request(options, (res) => {
            let data = '';
            res.on('data', chunk => data += chunk);
            res.on('end', () => {
                if (res.statusCode >= 200 && res.statusCode < 300) {
                    resolve(data);
                } else {
                    reject(new Error(`Server returned ${res.statusCode}: ${data}`));
                }
            });
        });

        req.on('error', reject);
        req.write(content);
        req.end();
    });
}

function sendQueryToServer(content, contentType) {
    return new Promise((resolve, reject) => {
        const options = {
            hostname: server,
            port: port,
            path: '/stores/' + _storeName + '/query',
            method: 'POST',
            headers: {
                'Content-Type': contentType,
                'Content-Length': Buffer.byteLength(content)
            }
        };

        const req = http.request(options, (res) => {
            let data = '';
            res.on('data', chunk => data += chunk);
            res.on('end', () => {
                if (res.statusCode >= 200 && res.statusCode < 300) {
                    resolve(data);
                } else {
                    reject(new Error(`Server returned ${res.statusCode}: ${data}`));
                }
            });
        });

        req.on('error', reject);
        req.write(content);
        req.end();
    });
}

function getWebviewContent(jsonResponse) {
    const htmlContent = `
    <!DOCTYPE html>
    <html>
    <head>
        <style>
            body { font-family: Arial, sans-serif; padding: 10px; }
            pre { background: #f4f4f4; padding: 10px; border-radius: 5px; }
        </style>
    </head>
    <body>
        <h1>Query Results</h1>
        <pre>${jsonResponse}</pre>
    </body>
    </html>
    `;
    return htmlContent;
}

let server = "localhost";
let port = 7642;
let username = "admin";
let password = "admin";

function activate(context) {
    const outputChannel = vscode.window.createOutputChannel("GraphLake");

    // Write a line of text
    outputChannel.appendLine("GraphLake is GO! - Dataplatform Solutions");

    // Show the output channel
    outputChannel.show(true);
    

    config = readConfig();
    if (config.server) {
        server = config.server;
        vscode.window.showInformationMessage(`Using API Endpoint: ${config.server}`);
    }

    if (config.port) {
        port = config.port;
        vscode.window.showInformationMessage(`Using Port: ${config.port}`);
    }

    if (config.username) {
        username = config.username;
        vscode.window.showInformationMessage(`Using Username: ${config.username}`);
    }

    if (config.password) {
        password = config.password;
        // vscode.window.showInformationMessage(`Using Password: ${config.password}`);
    }

    // data provider
    const dataProvider = new DataProvider();
    vscode.window.createTreeView("dataView", {
        treeDataProvider: dataProvider,
    });

    context.subscriptions.push(
        vscode.commands.registerCommand('graphlake.sendQueryToServer', async () => {
            const activeEditor = vscode.window.activeTextEditor;
            if (!activeEditor) {
                vscode.window.showErrorMessage("No active editor with a query file.");
                return;
            }
    
            const document = activeEditor.document;
            const fileExtension = document.fileName.split('.').pop();
            let contentType;
    
            // Determine the Content-Type based on the file extension
            if (fileExtension === 'js') {
                contentType = 'application/vnd.graphlake.x-embedded';
            } else if (fileExtension === 'rq') {
                contentType = 'application/sparql-query';
            } else if (fileExtension === 'ru') {
                contentType = 'application/sparql-update';
            } else {
                vscode.window.showErrorMessage("Unsupported file type for queries.");
                return;
            }
    
            const content = document.getText();
            try {
                vscode.window.withProgress(
                    {
                        location: vscode.ProgressLocation.Notification,
                        title: "Sending Query to Server...",
                        cancellable: false
                    },
                    async () => {
                        const response = await sendQueryToServer(content, contentType);
                        const panel = vscode.window.createWebviewPanel(
                            'queryResults',
                            'Query Results',
                            vscode.ViewColumn.Beside,
                            { enableScripts: true }
                        );
                        panel.webview.html = getWebviewContent(response);
                    }
                );
            } catch (error) {
                vscode.window.showErrorMessage(`Failed to send query: ${error.message}`);
            }
        })
    );

    // Create the "Graphs" and "Queries" views
    const graphsProvider = new GraphsProvider();
    vscode.window.createTreeView('graphsView', { treeDataProvider: graphsProvider });

    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.selectGraph", (selectedGraphName) => {
            _graphName = selectedGraphName; 
            vscode.window.showInformationMessage(`Switched to graph: "${selectedGraphName}"`);
        })
    );

    // Register the Create Graph command
    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.createGraph", async () => {
            const graphName = await vscode.window.showInputBox({
                placeHolder: "Enter the name of the new graph, enter to confirm, ESC to cancel",
                prompt: "Name your graph",
                validateInput: (value) => {
                    if (!value.trim()) {
                        return "Graph name cannot be empty.";
                    }
                    if (graphsProvider.graphs.includes(value.trim())) {
                        return `Graph "${value.trim()}" already exists.`;
                    }
                    return null;
                },
            });

            if (graphName) {
                graphsProvider.createGraph(graphName.trim());
                graphsProvider.loadGraphs();
                graphsProvider.refresh();
                vscode.window.showInformationMessage(`Graph "${graphName.trim()}" created.`);
            }
        })
    );

    
    const queriesProvider = new QueriesProvider();
    const queriesTreeView = vscode.window.createTreeView('queriesView', { treeDataProvider: queriesProvider });

    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.refreshQueries", () => {
            queriesProvider.refresh();
            vscode.window.showInformationMessage("Queries refreshed");
        })
    );

    // Handle visibility changes for the Queries view (optional)
    queriesTreeView.onDidChangeVisibility((event) => {
        if (event.visible) {
            queriesProvider.refresh();
        }
    });

    // Register the "Send Query to Server" command
    context.subscriptions.push(
        vscode.commands.registerCommand('graphlake.sendQueryToServer', async () => {
            const activeEditor = vscode.window.activeTextEditor;
            if (!activeEditor) {
                vscode.window.showErrorMessage("No active editor with a query file.");
                return;
            }

            const document = activeEditor.document;
            if (document.languageId !== 'javascript') {
                vscode.window.showErrorMessage("Only JavaScript files can be sent as queries.");
                return;
            }

            const content = document.getText();
            try {
                vscode.window.withProgress(
                    {
                        location: vscode.ProgressLocation.Notification,
                        title: "Sending Query to Server...",
                        cancellable: false
                    },
                    async () => {
                        const response = await sendQueryToServer(content);
                        const panel = vscode.window.createWebviewPanel(
                            'queryResults',
                            'Query Results',
                            vscode.ViewColumn.Beside,
                            { enableScripts: true }
                        );
                        panel.webview.html = getWebviewContent(response);
                    }
                );
            } catch (error) {
                vscode.window.showErrorMessage(`Failed to send query: ${error.message}`);
            }
        })
    );

    // Create the Status Bar Button
    createStatusBarButton();

    // Show/Hide the button based on active editor file type
    context.subscriptions.push(
        vscode.window.onDidChangeActiveTextEditor((editor) => {
            if (editor && editor.document.languageId === "javascript") {
                statusBarItem.show(); // Show button for .js files
            } else {
                statusBarItem.hide(); // Hide button for other files
            }
        })
    );

    // Handle initial state
    if (vscode.window.activeTextEditor?.document.languageId === "javascript") {
        statusBarItem.show();
    } else {
        statusBarItem.hide();
    }

    // Ensure the status bar item is disposed when the extension is deactivated
    context.subscriptions.push({
        dispose: () => {
            if (statusBarItem) {
                statusBarItem.dispose();
            }
        }
    });


    const storesProvider = new StoresProvider(graphsProvider);
    vscode.window.createTreeView("storesView", {
        treeDataProvider: storesProvider,
    });

    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.refreshStores", () => {
            storesProvider.loadStores();
            storesProvider.refresh();
            vscode.window.showInformationMessage("Stores refreshed");
        })
    );

    // Register the "Select Store" command
    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.selectStore", (storeName) => {
            _storeName = storeName;
            graphsProvider.loadGraphs();
            graphsProvider.refresh();
            vscode.window.showInformationMessage(`Switched to store: "${storeName}"`);
        })
    );

    // Register the "Create Store" command
    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.createStore", async () => {
            const storeName = await vscode.window.showInputBox({
                placeHolder: "Enter the name of the new store",
                prompt: "Create a new store",
                validateInput: (value) => {
                    if (!value.trim()) {
                        return "Store name cannot be empty.";
                    }
                    return null;
                },
            });

            if (storeName) {
                storesProvider.addStore(storeName.trim());
                vscode.window.showInformationMessage(`Store "${storeName.trim()}" created.`);
            }
        })
    );

    // Register the "Delete Store" command
    context.subscriptions.push(
        vscode.commands.registerCommand("graphlake.deleteStore", async (storeName) => {
            const confirm = await vscode.window.showWarningMessage(
                `Are you sure you want to delete the store "${storeName}"?`,
                { modal: true },
                "Yes"
            );

            if (confirm === "Yes") {
                storesProvider.deleteStore(storeName);
                vscode.window.showInformationMessage(`Store "${storeName}" deleted.`);
            }
        })
    );


}

function deactivate() {}

module.exports = {
    activate,
    deactivate
};